from unittest.mock import patch

from mau.parsers.text_parser import TextParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory

init_parser = init_parser_factory(TextParser)

_test = parser_test_factory(TextParser)


def test_collect_macro_arguments_single_argument():
    source = "(value1)"

    p = init_parser(source)
    assert p._collect_macro_args() == "value1"


def test_collect_macro_arguments_multiple_arguments():
    source = "(value1,value2)"

    p = init_parser(source)
    assert p._collect_macro_args() == "value1,value2"


def test_collect_macro_arguments_single_argument_with_quotes():
    source = '("value1")'

    p = init_parser(source)
    assert p._collect_macro_args() == '"value1"'


def test_collect_macro_arguments_single_argument_with_quotes_and_parenthesis():
    source = '("value1()")'

    p = init_parser(source)
    assert p._collect_macro_args() == '"value1()"'


def test_collect_macro_arguments_single_argument_with_parenthesis():
    source = "(value1())"

    p = init_parser(source)
    assert p._collect_macro_args() == "value1("


def test_collect_macro_arguments_multiple_argument_with_quotes_and_parenthesis():
    source = '("value1()",value2,value3)'

    p = init_parser(source)
    assert p._collect_macro_args() == '"value1()",value2,value3'


def test_macro():
    source = "[macroname](value1,value2)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "macro",
                    "name": "macroname",
                    "value": "value1",
                    "args": ["value1", "value2"],
                    "kwargs": {},
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_arguments_with_quotes():
    source = '[macroname]("value1,value2")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "macro",
                    "name": "macroname",
                    "value": "value1,value2",
                    "args": ["value1,value2"],
                    "kwargs": {},
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_named_arguments():
    source = "[macroname](name,arg1=value1)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "macro",
                    "name": "macroname",
                    "value": "name",
                    "args": ["name"],
                    "kwargs": {"arg1": "value1"},
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link():
    source = '[link](https://somedomain.org/the/path, "link text")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "link text",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_quotes():
    source = '[link](https://somedomain.org/the/path,"link, text")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "link, text",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_escaped_quotes():
    source = r'[link](https://somedomain.org/the/path,"link \"text\"")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": 'link "text"',
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_escaped_round_braces():
    source = '[link](https://somedomain.org/the/path,"link (text\\)")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "link (text)",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_escaped_round_braces_in_the_url():
    source = '[link](https://somedomain.org/the/path_(note\\),"link")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path_(note)",
                    "text": "link",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_quotes_and_round_braces_in_the_url():
    source = '[link]("https://somedomain.org/the/path_(note)","link")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path_(note)",
                    "text": "link",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_without_text():
    source = "[link](https://somedomain.org/the/path)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "https://somedomain.org/the/path",
                }
            ],
        }
    ]

    _test(source, expected)


def test_automatic_link():
    source = "This is https://somedomain.org/the/path text"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "This is "},
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "https://somedomain.org/the/path",
                },
                {"type": "text", "value": " text"},
            ],
        }
    ]

    _test(source, expected)


def test_automatic_link_ending_with_period():
    source = "The link is https://somedomain.org/the/path."

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "The link is "},
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "https://somedomain.org/the/path",
                },
                {"type": "text", "value": "."},
            ],
        }
    ]

    _test(source, expected)


def test_http_mentions():
    source = "Dangerous words: http and https"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Dangerous words: http and https"},
            ],
        }
    ]

    _test(source, expected)


def test_macro_image():
    source = "[image](/the/path.jpg)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": None,
                    "width": None,
                    "height": None,
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_image_with_alt_text():
    source = '[image](/the/path.jpg,"alt name")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": "alt name",
                    "width": None,
                    "height": None,
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_image_with_width_and_height():
    source = "[image](/the/path.jpg,width=120,height=120)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": None,
                    "width": "120",
                    "height": "120",
                }
            ],
        }
    ]

    _test(source, expected)


@patch("mau.parsers.text_parser.footnote_anchors")
def test_macro_footnote(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = "[footnote](Some text _and style_ and *more style* here)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "type": "sentence",
                            "content": [
                                {"type": "text", "value": "Some text "},
                                {
                                    "type": "style",
                                    "value": "underscore",
                                    "content": {
                                        "type": "sentence",
                                        "content": [
                                            {"type": "text", "value": "and style"},
                                        ],
                                    },
                                },
                                {"type": "text", "value": " and "},
                                {
                                    "type": "style",
                                    "value": "star",
                                    "content": {
                                        "type": "sentence",
                                        "content": [
                                            {"type": "text", "value": "more style"},
                                        ],
                                    },
                                },
                                {"type": "text", "value": " here"},
                            ],
                        }
                    ],
                }
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnote_defs) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text "},
                        {
                            "type": "style",
                            "value": "underscore",
                            "content": {
                                "type": "sentence",
                                "content": [
                                    {"type": "text", "value": "and style"},
                                ],
                            },
                        },
                        {"type": "text", "value": " and "},
                        {
                            "type": "style",
                            "value": "star",
                            "content": {
                                "type": "sentence",
                                "content": [
                                    {"type": "text", "value": "more style"},
                                ],
                            },
                        },
                        {"type": "text", "value": " here"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_can_start_with_different_number(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    p = init_parser(
        "[footnote](Some text _and style_ and *more style* here)",
        footnotes_start_with=6,
    )
    result = p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "footnote_ref",
                    "number": 6,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "content": [
                                {"type": "text", "value": "Some text "},
                                {
                                    "type": "style",
                                    "value": "underscore",
                                    "content": {
                                        "type": "sentence",
                                        "content": [
                                            {"type": "text", "value": "and style"},
                                        ],
                                    },
                                },
                                {"type": "text", "value": " and "},
                                {
                                    "type": "style",
                                    "value": "star",
                                    "content": {
                                        "type": "sentence",
                                        "content": [
                                            {"type": "text", "value": "more style"},
                                        ],
                                    },
                                },
                                {"type": "text", "value": " here"},
                            ],
                            "type": "sentence",
                        },
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_with_other_text_around(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = "text[footnote](Some text) other text"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "type": "sentence",
                            "content": [
                                {"type": "text", "value": "Some text"},
                            ],
                        }
                    ],
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnote_defs) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_includes_round_brakets(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = r'text[footnote]("Some code()") other text'

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "type": "sentence",
                            "content": [
                                {"type": "text", "value": "Some code()"},
                            ],
                        }
                    ],
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnote_defs) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some code()"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_commas_mau_v1(
    footnote_anchors_mock,
):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = r"text[footnote](Some text, that should not, be split) other text"

    _test = parser_test_factory(TextParser, v1_backward_compatibility=True)

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "type": "sentence",
                            "content": [
                                {
                                    "type": "text",
                                    "value": "Some text, that should not, be split",
                                },
                            ],
                        }
                    ],
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnote_defs) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {
                            "type": "text",
                            "value": "Some text, that should not, be split",
                        },
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_commas_mau_v2(
    footnote_anchors_mock,
):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = r'text[footnote]("Some text, that should not, be split") other text'

    _test = parser_test_factory(TextParser)

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                    "content": [
                        {
                            "type": "sentence",
                            "content": [
                                {
                                    "type": "text",
                                    "value": "Some text, that should not, be split",
                                },
                            ],
                        }
                    ],
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnote_defs) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {
                            "type": "text",
                            "value": "Some text, that should not, be split",
                        },
                    ],
                }
            ],
        }
    ]


def test_parse_macro_mailto():
    source = "[mailto](info@projectmau.org)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "mailto:info@projectmau.org",
                    "text": "info@projectmau.org",
                }
            ],
        }
    ]

    _test(source, expected)


def test_single_class():
    source = 'Some text [class]("text with that class", classname)'

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text "},
                {
                    "type": "class",
                    "classes": ["classname"],
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "text with that class"}],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_multiple_classes():
    source = 'Some text [class]("text with that class", "classname1,classname2")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text "},
                {
                    "type": "class",
                    "classes": ["classname1", "classname2"],
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "text with that class"}],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_parse_class_with_rich_text():
    source = '[class]("Some text with `verbatim words` and _styled ones_", classname)'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "class",
                    "classes": ["classname"],
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Some text with "},
                            {"type": "verbatim", "value": "verbatim words"},
                            {"type": "text", "value": " and "},
                            {
                                "type": "style",
                                "value": "underscore",
                                "content": {
                                    "type": "sentence",
                                    "content": [
                                        {"type": "text", "value": "styled ones"}
                                    ],
                                },
                            },
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)



import itertools

from mau.lexers.base_lexer import Token, TokenTypes, Literal
from mau.lexers.text_lexer import TextLexer
from mau.parsers.base_parser import BaseParser, ParserError
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.nodes import (
    WordNode,
    SentenceNode,
    TextNode,
    StyleNode,
    VerbatimNode,
    ClassNode,
    MacroNode,
    LinkNode,
    ImageNode,
    FootnoteRefNode,
    FootnoteDefNode,
)

# This is a simple map to keep track of the official
# name of styles introduced by special characters
MAP_STYLES = {"_": "underscore", "*": "star"}


def footnote_anchors(content):
    """
    Return the id of the reference and the id of the definition.
    """
    h = str(hash(content))[:8]
    return f"fr-{h}", f"fd-{h}"  # pragma: no cover


# The TextParser is a nested parser.
# The parsing always starts with parse_sentence
# and from there all components of the text are explored
class TextParser(BaseParser):
    def __init__(self, footnotes_start_with=1, v1_backward_compatibility=False):
        super().__init__()

        self.lexer = TextLexer()

        # This makes footnotes counting start from
        # the given number. It is needed because
        # there might be already footnotes in the
        # parent text to take into account.
        self.footnotes_start_with = footnotes_start_with

        # These are the footnotes found in this text
        self.footnote_defs = []

        # This enables backward compatibility with Mau 1.x
        # Footnotes didn't require quotes if they
        # contain commas. In Mau 2.x footnotes accept
        # arguments as the other macros so the text
        # should be wrapped in quotes.
        self.v1_backward_compatibility = v1_backward_compatibility

        # These are the nodes created by the parsing.
        self.nodes = []

        # This is used as a storage for macro arguments.
        self.argsparser = ArgumentsParser()

    def set_names_and_defaults(self, positional_names, default_values=None):
        """
        A wrapper for the method set_names_and_defaults() of ArgumentParser
        that raises a ParserError with a broader context.
        The argument parser is applied on the arguments only, so the error
        reporting is limited to that string, while the TextParser has knowledge
        of the whole line.
        """

        try:
            self.argsparser.set_names_and_defaults(positional_names, default_values)
        except ParserError as e:
            self.error(str(e))

    def parse_word(self):
        """
        Parse a single word.
        """
        return WordNode(self.get_token().value)

    def parse_style(self):
        """
        Parse a sentence surrounded by style markers.
        """

        # Get the style marker
        style = self.get_token(TokenTypes.LITERAL, check=lambda x: x in "*_").value

        # Get everything until the next marker
        content = self.parse_sentence(stop_tokens={Token(TokenTypes.LITERAL, style)})

        # Get the closing marker
        self.get_token(TokenTypes.LITERAL, style)

        return StyleNode(MAP_STYLES[style], content)

    def parse_escape(self):
        """
        Parse an escaped element.
        """

        # Drop the backslash
        self.get_token(TokenTypes.LITERAL, "\\")

        return WordNode(self.get_token().value)

    def parse_styled_text(self, stop_tokens=None):
        """
        Parse multiple possible elements: escapes, classes,
        macros, verbatim, styles, links, words.
        This is a helper for the function parse_sentence
        that takes into account all possible elements of
        syntax, stopping if the token is among the listed ones.
        """
        stop_tokens = stop_tokens or set()

        if self.peek_token() in stop_tokens:
            return None

        with self:
            return self.parse_escape()

        with self:
            return self.parse_class()

        with self:
            return self.parse_macro()

        with self:
            return self.parse_verbatim()

        with self:
            return self.parse_style()

        with self:
            return self.parse_link()

        return self.parse_word()

    def parse_sentence(self, stop_tokens=None):
        """
        Parse a sentence, which is made of multiple
        elements identified by parse_styled_text, until
        the EOF, the EOL, or a specific set of tokens
        passed as argument.
        """

        content = []
        stop_tokens = stop_tokens or set()
        stop_tokens = stop_tokens.union({Token(TokenTypes.EOF), Token(TokenTypes.EOL)})

        result = self.parse_styled_text(stop_tokens)
        while result is not None:
            content.append(result)
            result = self.parse_styled_text(stop_tokens)

        # Group consecutive WordNode nodes into a single TextNode
        grouped_nodes = []
        for key, group in itertools.groupby(content, lambda x: x.__class__ == WordNode):
            if key:
                text = "".join([n.value for n in group])
                grouped_nodes.append(TextNode(text))
            else:
                grouped_nodes.extend(list(group))

        return SentenceNode(content=grouped_nodes)

    def parse_verbatim(self):
        """
        Parse text in `verbatim`.
        """

        # Get the verbatim marker
        self.get_token(TokenTypes.LITERAL, "`")

        # Get the content tokens until the
        # next verbatim marker or EOL
        content = self.collect_join(
            [Token(TokenTypes.LITERAL, "`"), Token(TokenTypes.EOL)],
        )

        # Remove the closing marker
        self.get_token(TokenTypes.LITERAL, "`")

        return VerbatimNode(content)

    def parse_class(self):
        """
        Parse a class in the form [class]#content#.
        """

        # Get the opening square bracket
        self.get_token(TokenTypes.LITERAL, "[")

        # Get everything up to the closing bracket
        classes = self.collect_join(
            [Token(TokenTypes.LITERAL, "]"), Token(TokenTypes.EOL)]
        )

        # Get the closing bracket
        self.get_token(TokenTypes.LITERAL, "]")

        # Multiple classes are separated by commas
        classes = classes.split(",")

        # Get the opening number sign
        self.get_token(TokenTypes.LITERAL, "#")

        # Get the content of the class
        content = self.parse_sentence(stop_tokens={Token(TokenTypes.LITERAL, "#")})

        # Get the closing number sign
        self.get_token(TokenTypes.LITERAL, "#")

        return ClassNode(classes, content)

    def parse_macro_link(self):
        """
        Parse a link macro in the form [link](target, text).
        """

        # Assign names and get the arguments
        self.set_names_and_defaults(["target", "text"], {"text": None})
        args, kwargs = self.argsparser.get_arguments_and_reset()

        # Set the name of the first two unnamed arguments

        # Get the target as it can be used as default text
        target = kwargs.get("target")

        # We might omit text
        text = kwargs.get("text")
        if text is None:
            text = target

        return LinkNode(target, text)

    def parse_macro_mailto(self):
        """
        Parse a mailto macro in the form [mailto](email).
        """

        # Assign names and get the arguments
        self.set_names_and_defaults(["email"])
        args, kwargs = self.argsparser.get_arguments_and_reset()

        email = kwargs.get("email")
        target = f"mailto:{email}"

        return LinkNode(target, email)

    def parse_macro_class(self):
        """
        Parse a class macro in the form [class](text, "class1,class2,...").
        """

        # Assign names and get the arguments
        self.set_names_and_defaults(["text", "classes"])
        args, kwargs = self.argsparser.get_arguments_and_reset()

        # Parse the text
        pt = TextParser().analyse(kwargs.get("text"))

        # Text should return a single sentence node
        result = pt.nodes[0]

        # Multiple classes are separated by commas
        classes = kwargs.get("classes").split(",")

        return ClassNode(classes, result)

    def parse_macro_image(self):
        """
        Parse an inline image macro in the form
        [image](uri, alt_text, width, height).
        """

        # Assign names and get the arguments
        self.set_names_and_defaults(
            ["uri", "alt_text", "width", "height"],
            {"alt_text": None, "width": None, "height": None},
        )
        args, kwargs = self.argsparser.get_arguments_and_reset()

        return ImageNode(
            uri=kwargs.get("uri"),
            alt_text=kwargs.get("alt_text"),
            width=kwargs.get("width"),
            height=kwargs.get("height"),
        )

    def parse_macro_footnote(self, arguments):
        """
        Parse a footnote macro in the form
        [footnote](content).
        """

        if self.v1_backward_compatibility:
            self.argsparser.get_arguments_and_reset()
            footnote_text = arguments
        else:
            # Assign names and get the arguments
            self.set_names_and_defaults(["text"])
            args, kwargs = self.argsparser.get_arguments_and_reset()
            footnote_text = kwargs["text"]

        refanchor, defanchor = footnote_anchors(footnote_text)
        number = self.footnotes_start_with + len(self.footnote_defs)

        p = TextParser().analyse(footnote_text)

        self.footnote_defs.append(
            FootnoteDefNode(
                refanchor=refanchor, defanchor=defanchor, number=number, content=p.nodes
            )
        )
        return FootnoteRefNode(
            refanchor=refanchor, defanchor=defanchor, number=number, content=p.nodes
        )

    def _collect_macro_args(self):
        self.get_token(TokenTypes.LITERAL, "(")

        all_args = []

        while not (
            self.peek_token_is(TokenTypes.LITERAL, ")")
            or self.peek_token_is(TokenTypes.EOL)
        ):
            if self.peek_token_is(TokenTypes.LITERAL, '"'):
                self.get_token(TokenTypes.LITERAL, '"')

                value = self.collect_join(
                    stop_tokens=[Token(TokenTypes.LITERAL, '"'), Token(TokenTypes.EOL)],
                )

                self.get_token(TokenTypes.LITERAL, '"')

                arguments = f'"{value}"'
            else:
                arguments = self.collect_join(
                    stop_tokens=[
                        Token(TokenTypes.LITERAL, ","),
                        Token(TokenTypes.LITERAL, ")"),
                        Token(TokenTypes.EOL),
                    ],
                )

            all_args.append(arguments)

        self.get_token(TokenTypes.LITERAL, ")")

        return "".join(all_args)

    def parse_macro(self):
        self.get_token(TokenTypes.LITERAL, "[")
        macro_name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, "]")

        arguments = self._collect_macro_args()

        self.argsparser.analyse(arguments)

        if macro_name == "link":
            return self.parse_macro_link()
        elif macro_name == "mailto":
            return self.parse_macro_mailto()
        elif macro_name == "class":
            return self.parse_macro_class()
        elif macro_name == "image":
            return self.parse_macro_image()
        elif macro_name == "footnote":
            return self.parse_macro_footnote(arguments)

        # Get the arguments
        args, kwargs = self.argsparser.get_arguments_and_reset()

        return MacroNode(macro_name, args, kwargs)

    def parse_link(self):
        link = self.get_token(
            TokenTypes.TEXT,
            check=lambda x: x.startswith("http://") or x.startswith("https://"),
        ).value

        # Check the last character.
        # Common punctuation shouldn't be part of the link
        if link[-1] in [".", ","]:
            self.put_token(Literal(link[-1]))
            link = link[:-1]

        return LinkNode(link, link)

    def parse(self):
        self._save(self.parse_sentence())



import re

from mau.lexers.base_lexer import BaseLexer, TokenTypes


class TextLexer(BaseLexer):
    def _process_whitespace(self):
        regexp = re.compile(r"\ +")

        match = regexp.match(self._tail)

        if not match:
            return None

        return self._create_token_and_skip(TokenTypes.TEXT, " ")

    def _process_literal(self):
        if self._current_char not in '_*`{}()[]#\\"':
            return None

        return self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)

    def _process_text(self):
        regexp = re.compile(r'[^_*`{}()[\]#"\\ ]+')

        match = regexp.match(self._tail)

        if not match:  # pragma: no cover
            return None

        return self._create_token_and_skip(TokenTypes.TEXT, match.group())

    def _process_functions(self):
        return [
            self._process_eof,
            self._process_eol,
            self._process_whitespace,
            self._process_literal,
            self._process_text,
        ]



import pytest

from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_source():
    source = """
    [myblock, engine=source, language=somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_with_title():
    source = """
    . Title
    [myblock, engine=source, language=somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": {
                "content": [{"type": "text", "value": "Title"}],
                "type": "sentence",
            },
            "content": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_engine_source_without_language():
    source = """
    [myblock, engine=source]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "text",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_ignores_mau_syntax():
    source = """
    [myblock, engine=source]
    ----
    :answer:42
    The answer is {answer}
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "text",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": ":answer:42"},
                {"type": "text", "value": "The answer is {answer}"},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_respects_spaces_and_indentation():
    source = """
    [myblock, engine=source]
    ----
      //    This is a comment
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "text",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "  //    This is a comment"},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callouts():
    source = """
    [myblock, engine=source, language=somelang, callouts=":"]
    ----
    import sys
    import os:3:
    
    print(os.environ["HOME"]):6:
    ----
    3: This is an import
    6: Environment variables are paramount
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "3", 3: "6"},
                    "contents": {
                        "3": "This is an import",
                        "6": "Environment variables are paramount",
                    },
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callouts_possible_clash():
    source = """
    [myblock, engine=source, language=somelang, callouts=":"]
    ----
    import sys
    import: os:3:
    
    print(os.environ["HOME"]):6:
    ----
    3: This is an import
    6: Environment variables are paramount
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "3", 3: "6"},
                    "contents": {
                        "3": "This is an import",
                        "6": "Environment variables are paramount",
                    },
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import: os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callouts_default_delimiter():
    source = """
    [myblock, engine=source, language=somelang]
    ----
    import sys
    import os:3:
    
    print(os.environ["HOME"]):6:
    ----
    3: This is an import
    6: Environment variables are paramount
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "3", 3: "6"},
                    "contents": {
                        "3": "This is an import",
                        "6": "Environment variables are paramount",
                    },
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callouts_custom_delimiter():
    source = """
    [myblock, engine=source, language=somelang, callouts="|"]
    ----
    import sys
    import os|3|

    print(os.environ["HOME"])|6|
    ----
    3: This is an import
    6: Environment variables are paramount
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "3", 3: "6"},
                    "contents": {
                        "3": "This is an import",
                        "6": "Environment variables are paramount",
                    },
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callout_cannot_be_found():
    p = init_parser(
        dedent(
            """
            [myblock, engine=source, language=somelang, callouts="|"]
            ----
            import sys
            import os

            print(os.environ["HOME"])|6|
            ----
            3: This is an import
            6: Environment variables are paramount
            """
        )
    )

    with pytest.raises(ParserError):
        p.parse()


def test_source_callouts_non_numeric_labels():
    source = """
    [myblock, engine=source, language=somelang, callouts="|"]
    ----
    import sys
    import os|step1|
    
    print(os.environ["HOME"])|step2|
    ----
    step1: This is an import
    step2: Environment variables are paramount
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "step1", 3: "step2"},
                    "contents": {
                        "step1": "This is an import",
                        "step2": "Environment variables are paramount",
                    },
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_callout_wrong_format():
    p = init_parser(
        dedent(
            """
            [myblock, engine=source, language=somelang, callouts=":"]
            ----
            import sys
            import os:3:

            print(os.environ["HOME"]):6:
            ----
            3 This is an import
            6: Environment variables are paramount
            """
        )
    )

    with pytest.raises(ParserError):
        p.parse()


def test_source_callouts_without_definition():
    source = """
    [myblock, engine=source, language=somelang, callouts=":"]
    ----
    import sys
    import os:3:
    
    print(os.environ["HOME"]):6:
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "somelang",
                "callouts": {
                    "markers": {1: "3", 3: "6"},
                    "contents": {},
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_highlights():
    source = """
    [myblock, engine=source, language=somelang]
    ----
    import sys
    import os:@:
    
    print(os.environ["HOME"]):@:
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [1, 3],
                "language": "somelang",
                "callouts": {
                    "markers": {},
                    "contents": {},
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_source_highlights_custom_marker():
    source = """
    [myblock, engine=source, language=somelang, highlight="#"]
    ----
    import sys
    import os:#:
    
    print(os.environ["HOME"]):#:
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "myblock",
            "args": [],
            "kwargs": {
                "highlights": [1, 3],
                "language": "somelang",
                "callouts": {
                    "markers": {},
                    "contents": {},
                },
            },
            "secondary_content": [],
            "title": None,
            "content": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)


def test_engine_source_language_is_reset():
    source = """
    [source, python]
    ----
    ----

    [source]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "source",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "python",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
        {
            "type": "block",
            "blocktype": "source",
            "args": [],
            "kwargs": {
                "highlights": [],
                "language": "text",
                "callouts": {"markers": {}, "contents": {}},
            },
            "secondary_content": [],
            "title": None,
            "content": [],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
        },
    ]

    _test(source, expected)



import pytest

from unittest.mock import patch

from mau.parsers.nodes import DocumentNode, ContainerNode
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser
from mau.visitors.html_visitor import HTMLVisitor

from tests.helpers import (
    dedent,
    init_parser_factory,
    init_ast_factory,
    ast_test_factory,
    remove_indentation,
    visitlist_factory,
)

init_parser = init_parser_factory(MainParser)

init_ast = init_ast_factory(MainParser)

_test = ast_test_factory(MainParser, HTMLVisitor)

visitlist = visitlist_factory(HTMLVisitor)


def test_text():
    v = HTMLVisitor()

    output = v.visit(
        {
            "type": "text",
            "value": "This is text",
        }
    )

    assert output == "This is text"


def test_sentence():
    v = HTMLVisitor()

    output = v.visit(
        {"type": "sentence", "content": [{"type": "text", "value": "This is text"}]}
    )

    assert output == "This is text"


def test_paragraph():
    source = "This is text"

    expected = ["<p>This is text</p>"]

    _test(source, expected)


def test_horizontal_rule():
    source = "---"

    expected = ["<hr>"]

    _test(source, expected)


def test_style_underscore():
    source = "_This is text_"

    expected = ["<p><em>This is text</em></p>"]

    _test(source, expected)


def test_text_that_contains_styles():
    source = "This _is_ text"

    expected = ["<p>This <em>is</em> text</p>"]

    _test(source, expected)


def test_style_star():
    source = "*This is text*"

    expected = ["<p><strong>This is text</strong></p>"]

    _test(source, expected)


def test_style_verbatim():
    source = "`This is text`"

    expected = ["<p><code>This is text</code></p>"]

    _test(source, expected)


def test_class_old_style():
    source = "[classname]#text with that class#"

    expected = ['<p><span class="classname">text with that class</span></p>']

    _test(source, expected)


def test_class_new_style():
    source = '[class]("text with classes", "class1,class2")'

    expected = ['<p><span class="class1 class2">text with classes</span></p>']

    _test(source, expected)


def test_link():
    source = '[link](https://somedomain.org/the/path,"link text")'

    expected = ['<p><a href="https://somedomain.org/the/path">link text</a></p>']

    _test(source, expected)


def test_inline_image():
    source = "[image](/the/path.jpg)"

    expected = ['<p><span class="image"><img src="/the/path.jpg"></span></p>']

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_header_level_3(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = "=== Title of the section"

    expected = ['<h3 id="XXXXXX">Title of the section</h3>']

    _test(source, expected)


def test_quote():
    source = dedent(
        """
        [quote,"Star Wars, 1977"]
        ----
        Learn about the Force, Luke.
        ----
        """
    )

    expected = [
        remove_indentation(
            """
            <blockquote>
              <p>Learn about the Force, Luke.</p>
              <cite>Star Wars, 1977</cite>
            </blockquote>
            """
        )
    ]

    _test(source, expected)


def test_quote_attribution_in_secondary_content_supports_styles():
    source = dedent(
        """
        [quote]
        ----
        Learn about the Force, Luke.
        ----
        _Star Wars_, 1977
        """
    )

    expected = [
        remove_indentation(
            """
            <blockquote>
              <p>Learn about the Force, Luke.</p>
              <cite><p><em>Star Wars</em>, 1977</p></cite>
            </blockquote>
            """
        )
    ]

    _test(source, expected)


def test_admonition():
    source = dedent(
        """
        [admonition,aclass,anicon,alabel]
        ----
        This is my test admonition
        ----
        """
    )

    expected = [
        remove_indentation(
            """
            <div class="admonition aclass">
              <i class="anicon"></i>
              <div class="content">
                <div class="title">alabel</div>
                <div>
                  <p>This is my test admonition</p>
                </div>
              </div>
            </div>
            """
        )
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_block(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = dedent(
        """
        [sometype]
        ++++
        = A block

        This contains headers, paragraphs and blocks

        ++++
        """
    )

    expected = [
        """<div class="sometype"><div class="content"><h1 id="XXXXXX">A block</h1>\n<p>This contains headers, paragraphs and blocks</p></div></div>"""
    ]

    _test(source, expected)


def test_basic_block_attributes():
    source = dedent(
        """
        [sometype,mayattr1=value1]
        ++++
        This contains headers, paragraphs and blocks
        ++++
        """
    )

    expected = [
        remove_indentation(
            """
            <div class="sometype">
              <div class="content">
                <p>This contains headers, paragraphs and blocks</p>
              </div>
            </div>
            """
        )
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_basic_block_title(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = dedent(
        """
        . Title
        [sometype]
        ++++
        = A block

        This contains headers, paragraphs and blocks

        ++++
        """
    )

    expected = [
        """<div class="sometype"><div class="title">Title</div><div class="content"><h1 id="XXXXXX">A block</h1>\n<p>This contains headers, paragraphs and blocks</p></div></div>"""
    ]

    _test(source, expected)


def test_document():
    parser = init_parser(
        dedent(
            """
            This is text
            """
        )
    )
    parser.parse()

    document = DocumentNode(parser.nodes)
    node = document.asdict()

    result = visitlist([node])

    assert result == [
        remove_indentation(
            """
            <html>
              <head>
              </head>
              <body>
                <p>This is text</p>
              </body>
            </html>
            """
        )
    ]


def test_container():
    parser = init_parser(
        dedent(
            """
            This is text
            """
        )
    )
    parser.parse()

    document = ContainerNode(parser.nodes)
    node = document.asdict()

    result = visitlist([node])

    assert result == [
        remove_indentation(
            """
            <p>This is text</p>
            """
        )
    ]


def test_list_unordered():
    source = dedent(
        """
        * Item 1
        ** Item 1.1
        * Item 2
        """
    )

    expected = [
        remove_indentation(
            """
            <ul>
              <li>Item 1
                <ul>
                  <li>Item 1.1</li>
                </ul>
              </li>
              <li>Item 2</li>
            </ul>
            """
        )
    ]

    _test(source, expected)


def test_list_ordered():
    source = dedent(
        """
        # Item 1
        ## Item 1.1
        # Item 2
        """
    )

    expected = [
        remove_indentation(
            """
            <ol>
              <li>Item 1
                <ol>
                  <li>Item 1.1</li>
                </ol>
              </li>
              <li>Item 2</li>
            </ol>
            """
        )
    ]

    _test(source, expected)


def test_list_mixed():
    source = dedent(
        """
        # Item 1
        ** Item 1.1
        # Item 2
        """
    )

    expected = [
        remove_indentation(
            """
            <ol>
              <li>Item 1
                <ul>
                  <li>Item 1.1</li>
                </ul>
              </li>
              <li>Item 2</li>
            </ol>
            """
        )
    ]

    _test(source, expected)


def test_include_image():
    source = "<< image:/path/to/it.jpg"

    expected = [
        remove_indentation(
            """
            <div class="imageblock">
              <div class="content">
                <img src="/path/to/it.jpg" />
              </div>
            </div>
            """
        )
    ]

    _test(source, expected)


def test_include_image_alt_text():
    source = dedent(
        """
        [alt_text="This is a beautiful image"]
        << image:/path/to/it.jpg
        """
    )

    expected = [
        remove_indentation(
            """
            <div class="imageblock">
              <div class="content">
                <img src="/path/to/it.jpg" alt="This is a beautiful image" />
              </div>
            </div>
            """
        )
    ]

    _test(source, expected)


def test_include_image_title():
    source = dedent(
        """
        . A nice _caption_
        << image:/path/to/it.jpg
        """
    )

    expected = [
        remove_indentation(
            """
            <div class="imageblock">
              <div class="content">
                <img src="/path/to/it.jpg" />
                <div class="title">A nice <em>caption</em></div>
              </div>
            </div>
            """
        )
    ]

    _test(source, expected)


def test_macro():
    source = "[unknown](This is an unknown macro)"

    expected = ["<p></p>"]

    _test(source, expected)


def test_macro_template():
    v = HTMLVisitor()
    v.default_templates["macro.html"] = "whatever"

    ast = init_ast("[unknown](This is an unknown macro)")

    output = [v.visit(node) for node in ast]

    assert output == ["<p>whatever</p>"]


def test_macro_uses_specific_template():
    v = HTMLVisitor()
    v.default_templates["macro.html"] = "whatever"
    v.default_templates["macro-unknown.html"] = "the right one"

    ast = init_ast("[unknown](This is an unknown macro)")

    output = [v.visit(node) for node in ast]

    assert output == ["<p>the right one</p>"]


def test_macro_provides_args_and_kwargs():
    v = HTMLVisitor()
    v.default_templates["macro.html"] = "whatever"
    v.default_templates["macro-unknown.html"] = "{{ args[0] }} - {{ kwargs.name}}"

    ast = init_ast('[unknown]("some text", name=mau)')

    output = [v.visit(node) for node in ast]

    assert output == ["<p>some text - mau</p>"]


def test_macro_alias_value_for_first_unnamed_arg():
    v = HTMLVisitor()
    v.default_templates["macro.html"] = "whatever"
    v.default_templates["macro-unknown.html"] = "{{ value }} - {{ kwargs.name}}"

    ast = init_ast('[unknown]("some text", name=mau)')

    output = [v.visit(node) for node in ast]

    assert output == ["<p>some text - mau</p>"]


def test_block_engine_raw():
    source = dedent(
        """
        [block, engine=raw]
        ----
        <div class="someclass">
          <p>This is HTML</>
        </div>
        ----
        """
    )

    expected = [
        """<div class="block"><div class="content"><div class="someclass">\n  <p>This is HTML</>\n</div></div></div>"""
    ]

    _test(source, expected)


def test_block_engine_default():
    source = dedent(
        """
        = Header out

        [block, engine=default]
        ----
        = Header in

        ::toc:
        ----
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == [
        '<h1 id="header-out">Header out</h1>',
        '<div class="block"><div class="content"><h1 id="header-in">Header in</h1>\n<div><ul><li><a href="#header-out">Header out</a></li><li><a href="#header-in">Header in</a></li></ul></div></div></div>',
    ]


def test_block_engine_mau():
    source = dedent(
        """
        = Header out

        [block, engine=mau]
        ----
        = Header in

        ::toc:
        ----
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == [
        '<h1 id="header-out">Header out</h1>',
        '<div class="block"><div class="content"><h1 id="header-in">Header in</h1><div><ul><li><a href="#header-in">Header in</a></li></ul></div></div></div>',
    ]


def test_block_template():
    v = HTMLVisitor()
    v.default_templates["block.html"] = "whatever"

    ast = init_ast(
        dedent(
            """
            [block]
            ----
            Primary content
            ----
            Secondary content
            """
        )
    )

    output = [v.visit(node) for node in ast]

    assert output == ["whatever"]


def test_block_uses_specific_template():
    v = HTMLVisitor()
    v.default_templates["block.html"] = "whatever"
    v.default_templates["block-unknown.html"] = "the right one"

    ast = init_ast(
        dedent(
            """
            [unknown]
            ----
            Primary content
            ----
            Secondary content
            """
        )
    )

    output = [v.visit(node) for node in ast]

    assert output == ["the right one"]


def test_block_supports_kwargs():
    v = HTMLVisitor()
    v.default_templates["block.html"] = "whatever"
    v.default_templates[
        "block-unknown.html"
    ] = "{{ kwargs.name }} - {{ kwargs.colour }}"

    ast = init_ast(
        dedent(
            """
            [unknown, name=myblock, colour=green]
            ----
            ----
            """
        )
    )

    output = [v.visit(node) for node in ast]

    assert output == ["myblock - green"]


def test_block_does_not_support_args():
    v = HTMLVisitor()
    v.default_templates["block.html"] = "whatever"
    v.default_templates["block-unknown.html"] = ""

    with pytest.raises(ParserError):
        ast = init_ast(
            dedent(
                """
                [unknown, value1, value2]
                ----
                Primary content
                ----
                Secondary content
                """
            )
        )


@patch("mau.parsers.main_parser.header_anchor")
def test_command_consumes_args(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = dedent(
        """
        ::unknown:unused1,unuse2=value2

        [sometype]
        ++++
        = A block

        This contains headers, paragraphs and blocks

        ++++
        """
    )

    expected = [
        "",
        """<div class="sometype"><div class="content"><h1 id="XXXXXX">A block</h1>\n<p>This contains headers, paragraphs and blocks</p></div></div>""",
    ]

    _test(source, expected)



from unittest.mock import patch

from mau.parsers import nodes
from mau.parsers.main_parser import MainParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            == Header 1.1
            == Header 1.2
            = Header 2
            == Header 2.1
            === Header 2.1.1
            """
        )
    )
    p.parse()

    assert p.toc.asdict() == {
        "type": "toc",
        "entries": [
            {
                "type": "toc_entry",
                "header": {
                    "level": 1,
                    "value": "Header 1",
                    "anchor": "Header 1-XXXXXX",
                    "tags": [],
                    "kwargs": {},
                    "type": "header",
                },
                "children": [
                    {
                        "type": "toc_entry",
                        "header": {
                            "level": 2,
                            "value": "Header 1.1",
                            "anchor": "Header 1.1-XXXXXX",
                            "tags": [],
                            "kwargs": {},
                            "type": "header",
                        },
                        "children": [],
                    },
                    {
                        "type": "toc_entry",
                        "header": {
                            "level": 2,
                            "value": "Header 1.2",
                            "anchor": "Header 1.2-XXXXXX",
                            "tags": [],
                            "kwargs": {},
                            "type": "header",
                        },
                        "children": [],
                    },
                ],
            },
            {
                "type": "toc_entry",
                "header": {
                    "level": 1,
                    "value": "Header 2",
                    "anchor": "Header 2-XXXXXX",
                    "tags": [],
                    "kwargs": {},
                    "type": "header",
                },
                "children": [
                    {
                        "type": "toc_entry",
                        "header": {
                            "level": 2,
                            "value": "Header 2.1",
                            "anchor": "Header 2.1-XXXXXX",
                            "tags": [],
                            "kwargs": {},
                            "type": "header",
                        },
                        "children": [
                            {
                                "type": "toc_entry",
                                "header": {
                                    "level": 3,
                                    "value": "Header 2.1.1",
                                    "anchor": "Header 2.1.1-XXXXXX",
                                    "tags": [],
                                    "kwargs": {},
                                    "type": "header",
                                },
                                "children": [],
                            },
                        ],
                    },
                ],
            },
        ],
    }


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc_ignore_headers(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            ==! Header 1.1
            ==! Header 1.2
            = Header 2
            ==! Header 2.1
            ===! Header 2.1.1
            """
        )
    )
    p.parse()

    assert p.toc.asdict() == {
        "type": "toc",
        "entries": [
            {
                "type": "toc_entry",
                "header": {
                    "level": 1,
                    "value": "Header 1",
                    "anchor": "Header 1-XXXXXX",
                    "tags": [],
                    "kwargs": {},
                    "type": "header",
                },
                "children": [],
            },
            {
                "type": "toc_entry",
                "header": {
                    "level": 1,
                    "value": "Header 2",
                    "anchor": "Header 2-XXXXXX",
                    "tags": [],
                    "kwargs": {},
                    "type": "header",
                },
                "children": [],
            },
        ],
    }


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc_orphan_nodes(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            === Header 1.1.1
            == Header 1.2
            """
        )
    )
    p.parse()

    assert p.toc.asdict() == {
        "type": "toc",
        "entries": [
            {
                "type": "toc_entry",
                "header": {
                    "level": 1,
                    "value": "Header 1",
                    "anchor": "Header 1-XXXXXX",
                    "tags": [],
                    "kwargs": {},
                    "type": "header",
                },
                "children": [
                    {
                        "type": "toc_entry",
                        "header": {
                            "level": 3,
                            "value": "Header 1.1.1",
                            "anchor": "Header 1.1.1-XXXXXX",
                            "tags": [],
                            "kwargs": {},
                            "type": "header",
                        },
                        "children": [],
                    },
                    {
                        "type": "toc_entry",
                        "header": {
                            "level": 2,
                            "value": "Header 1.2",
                            "anchor": "Header 1.2-XXXXXX",
                            "tags": [],
                            "kwargs": {},
                            "type": "header",
                        },
                        "children": [],
                    },
                ],
            },
        ],
    }


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_not_in_toc(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    =! Title of the section
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Title of the section",
            "level": 1,
            "anchor": "XXXXXX",
        }
    ]

    p = _test(source, expected)

    assert p.toc.asdict() == {
        "type": "toc",
        "entries": [],
    }



from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(TextParser)

_test = parser_test_factory(TextParser)


def test_empty_text():
    source = ""

    expected = [
        {
            "type": "sentence",
            "content": [],
        }
    ]

    _test(source, expected)


def test_multiple_words():
    source = "Many different words"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Many different words"},
            ],
        }
    ]

    _test(source, expected)


def test_underscore():
    source = "_Some text_"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "Some text"}],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_star():
    source = "*Some text*"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "style",
                    "value": "star",
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "Some text"}],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_style_within_style():
    source = "_*Words with two styles*_"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "style",
                                "value": "star",
                                "content": {
                                    "type": "sentence",
                                    "content": [
                                        {
                                            "type": "text",
                                            "value": "Words with two styles",
                                        }
                                    ],
                                },
                            }
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_paragraph_double_style_cancels_itself():
    source = "__Text__"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {"type": "sentence", "content": []},
                },
                {"type": "text", "value": "Text"},
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {"type": "sentence", "content": []},
                },
            ],
        }
    ]

    _test(source, expected)


def test_text_and_styles():
    source = "Some text _and style_ and *more style* here"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text "},
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "and style"},
                        ],
                    },
                },
                {"type": "text", "value": " and "},
                {
                    "type": "style",
                    "value": "star",
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "more style"},
                        ],
                    },
                },
                {"type": "text", "value": " here"},
            ],
        }
    ]

    _test(source, expected)


def test_parse_style_open():
    source = "_Text"

    expected = [
        {
            "type": "sentence",
            "content": [{"type": "text", "value": "_Text"}],
        },
    ]

    _test(source, expected)


def test_verbatim():
    source = "`Many different words`"

    expected = [
        {
            "type": "sentence",
            "content": [{"type": "verbatim", "value": "Many different words"}],
        }
    ]

    _test(source, expected)


def test_verbatim_backtick():
    source = r"`\``"

    expected = [
        {
            "type": "sentence",
            "content": [{"type": "verbatim", "value": "`"}],
        }
    ]

    _test(source, expected)


def test_verbatim_ignore_style():
    source = "`_Many different words_`"

    expected = [
        {
            "type": "sentence",
            "content": [{"type": "verbatim", "value": "_Many different words_"}],
        }
    ]

    _test(source, expected)


def test_verbatim_open():
    source = "`Many different words"

    expected = [
        {
            "type": "sentence",
            "content": [{"type": "text", "value": "`Many different words"}],
        }
    ]

    _test(source, expected)


def test_verbatim_and_style():
    source = "Some text with `verbatim words` and _styled ones_"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text with "},
                {"type": "verbatim", "value": "verbatim words"},
                {"type": "text", "value": " and "},
                {
                    "type": "style",
                    "value": "underscore",
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "styled ones"}],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_parse_class_with_rich_text():
    source = "[classname]#Some text with `verbatim words` and _styled ones_#"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "class",
                    "classes": ["classname"],
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Some text with "},
                            {"type": "verbatim", "value": "verbatim words"},
                            {"type": "text", "value": " and "},
                            {
                                "type": "style",
                                "value": "underscore",
                                "content": {
                                    "type": "sentence",
                                    "content": [
                                        {"type": "text", "value": "styled ones"}
                                    ],
                                },
                            },
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_single_class():
    source = "Some text [classname]#text with that class#"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text "},
                {
                    "type": "class",
                    "classes": ["classname"],
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "text with that class"}],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_multiple_classes():
    source = "Some text [classname1,classname2]#text with that class#"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Some text "},
                {
                    "type": "class",
                    "classes": ["classname1", "classname2"],
                    "content": {
                        "type": "sentence",
                        "content": [{"type": "text", "value": "text with that class"}],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_class_with_verbatim_and_class_literal():
    source = "[classname]#`verbatim #`#"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "class",
                    "classes": ["classname"],
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "verbatim", "value": "verbatim #"},
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_square_brackets():
    source = "This contains [ and ] and [this]"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "This contains [ and ] and [this]"},
            ],
        }
    ]

    _test(source, expected)


def test_escapes():
    source = r"This contains \[this\]#that looks like a class#"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "text",
                    "value": "This contains [this]#that looks like a class#",
                },
            ],
        }
    ]

    _test(source, expected)



import pytest

from mau.parsers import nodes
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser, EngineError

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_attributes_block():
    source = """
    [blocktype,myattr1=value1]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"myattr1": "value1"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


def test_attributes_can_contain_variables():
    source = """
    :value:42

    [blocktype,myattr1={value}]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"myattr1": "42"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


def test_parse_block_title_and_attributes():
    source = """
    .Just a title
    [blocktype, name1=value1, name2=value2]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "blocktype",
            "content": [],
            "secondary_content": [],
            "title": {
                "content": [{"type": "text", "value": "Just a title"}],
                "type": "sentence",
            },
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "args": [],
            "kwargs": {"name1": "value1", "name2": "value2"},
        },
    ]

    _test(source, expected)


def test_parse_block_title_and_attributes_are_reset():
    source = """
    .Just a title
    [blocktype1, name1=value1, name2=value2]
    ----
    ----

    [blocktype2]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "blocktype1",
            "content": [],
            "secondary_content": [],
            "title": {
                "content": [{"type": "text", "value": "Just a title"}],
                "type": "sentence",
            },
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "args": [],
            "kwargs": {"name1": "value1", "name2": "value2"},
        },
        {
            "type": "block",
            "blocktype": "blocktype2",
            "content": [],
            "secondary_content": [],
            "title": None,
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_block_classes_single_class():
    source = """
    [blocktype,classes=cls1]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {},
            "secondary_content": [],
            "classes": ["cls1"],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_classes_multiple_classes():
    source = """
    [blocktype,classes="cls1,cls2"]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {},
            "secondary_content": [],
            "classes": ["cls1", "cls2"],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_engine():
    source = """
    [blocktype,engine=someengine]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {},
            "secondary_content": [],
            "classes": [],
            "engine": "someengine",
            "title": None,
            "content": [],
        },
    ]

    with pytest.raises(EngineError):
        _test(source, expected)


def test_block_mau_has_no_external_variables():
    source = """
    :answer:42
    [block, engine=mau]
    ----
    The answer is {answer}.
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "block",
            "content": [{"type": "text", "value": "The answer is {answer}."}],
            "secondary_content": [],
            "title": None,
            "classes": [],
            "engine": "mau",
            "preprocessor": "none",
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_block_raw_engine():
    source = """
    [block, engine=raw]
    ----
    Raw content
    on multiple lines
    ----
    Secondary content
    on multiple lines too
    """

    expected = [
        {
            "type": "block",
            "blocktype": "block",
            "content": [
                {"type": "text", "value": "Raw content"},
                {"type": "text", "value": "on multiple lines"},
            ],
            "secondary_content": [
                {"type": "text", "value": "Secondary content"},
                {"type": "text", "value": "on multiple lines too"},
            ],
            "title": None,
            "classes": [],
            "engine": "raw",
            "preprocessor": "none",
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_block_default_engine_adds_headers_to_global_toc():
    source = """
    = Global header

    [someblock]
    ----
    = Block header
    ----
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Global header",
            "level": 1,
            "anchor": "global-header",
        },
        {
            "args": [],
            "blocktype": "someblock",
            "content": [
                {
                    "anchor": "block-header",
                    "kwargs": {},
                    "level": 1,
                    "tags": [],
                    "type": "header",
                    "value": "Block header",
                },
            ],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "kwargs": {},
            "secondary_content": [],
            "title": None,
            "type": "block",
        },
    ]

    p = _test(source, expected)

    assert p.headers == [
        nodes.HeaderNode("Global header", 1, "global-header"),
        nodes.HeaderNode("Block header", 1, "block-header"),
    ]


def test_block_positive_condition_matches():
    source = """
    :render:yes

    [block, condition="if:render:yes"]
    ----
    This is a paragraph.
    ----
    """

    expected = [
        {
            "args": [],
            "blocktype": "block",
            "classes": [],
            "content": [
                {
                    "args": [],
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a paragraph.",
                            },
                        ],
                        "type": "sentence",
                    },
                    "kwargs": {},
                    "type": "paragraph",
                },
            ],
            "engine": "default",
            "preprocessor": "none",
            "kwargs": {},
            "secondary_content": [],
            "title": None,
            "type": "block",
        },
    ]

    _test(source, expected)


def test_block_positive_condition_doesnt_match():
    source = """
    :render:no

    [block, condition="if:render:yes"]
    ----
    This is a paragraph.
    ----
    """

    expected = []

    _test(source, expected)


def test_block_negative_condition_matches():
    source = """
    :render:yes

    [block, condition="ifnot:render:no"]
    ----
    This is a paragraph.
    ----
    """

    expected = [
        {
            "args": [],
            "blocktype": "block",
            "classes": [],
            "content": [
                {
                    "args": [],
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a paragraph.",
                            },
                        ],
                        "type": "sentence",
                    },
                    "kwargs": {},
                    "type": "paragraph",
                },
            ],
            "engine": "default",
            "preprocessor": "none",
            "kwargs": {},
            "secondary_content": [],
            "title": None,
            "type": "block",
        },
    ]

    _test(source, expected)


def test_block_negative_condition_doesnt_match():
    source = """
    :render:no

    [block, condition="ifnot:render:no"]
    ----
    This is a paragraph.
    ----
    """

    expected = []

    _test(source, expected)


def test_block_condition_accepts_booleans():
    source = """
    :render:

    [block, condition="ifnot:render:"]
    ----
    This is a paragraph.
    ----
    """

    expected = []

    _test(source, expected)


def test_block_condition_raises_exception():
    source = """
    [block, condition="if:render"]
    ----
    This is a paragraph.
    ----
    """

    expected = []

    with pytest.raises(ParserError):
        _test(source, expected)


def test_command_defblock():
    p = init_parser("::defblock:alias, myblock, language=python, engine=source")
    p.parse()

    assert p.block_aliases["alias"] == "myblock"
    assert p.block_defaults["myblock"] == {"language": "python", "engine": "source"}


def test_command_defblock_backward_compatible_source_can_be_overridden():
    p = init_parser("::defblock:source, source, language=python, engine=source")
    p.parse()

    assert p.block_aliases["source"] == "source"
    assert p.block_defaults["source"] == {"language": "python", "engine": "source"}


def test_command_defblock_no_args():
    p = init_parser("::defblock:")

    with pytest.raises(ParserError):
        p.parse()


def test_command_defblock_single_arg():
    p = init_parser("::defblock:source")

    with pytest.raises(ParserError):
        p.parse()


def test_block_definitions_are_used():
    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"name1": "value1", "name2": "value2"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_local_kwargs_overwrite_defined_ones():
    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias, name1=value99]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"name1": "value99", "name2": "value2"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_local_args_are_used():
    source = """
    ::defblock:alias, blocktype, name1=value1, name2=value2

    [alias, attr1, name1=value99]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": ["attr1"],
            "blocktype": "blocktype",
            "kwargs": {"name1": "value99", "name2": "value2"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    with pytest.raises(ParserError):
        _test(source, expected)


def test_block_definitions_default_source_has_engine_source():
    source = """
    [source]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "source",
            "kwargs": {
                "callouts": {"contents": {}, "markers": {}},
                "highlights": [],
                "language": "text",
            },
            "secondary_content": [],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_default_source_language():
    source = """
    [source, python]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "source",
            "kwargs": {
                "callouts": {"contents": {}, "markers": {}},
                "highlights": [],
                "language": "python",
            },
            "secondary_content": [],
            "classes": [],
            "engine": "source",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_unnamed_args_are_used_as_names():
    source = """
    ::defblock:alias, blocktype, attr1, attr2

    [alias, value1, value2]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"attr1": "value1", "attr2": "value2"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_unnamed_and_named_args():
    source = """
    ::defblock:alias, blocktype, attr1, attr2, attr3=value3

    [alias, value1, value2]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"attr1": "value1", "attr2": "value2", "attr3": "value3"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    _test(source, expected)


def test_block_definitions_no_values_for_unnamed_args():
    source = """
    ::defblock:alias, blocktype, attr1, attr2, attr3=value3

    [alias]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"attr3": "value3"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    with pytest.raises(ParserError):
        _test(source, expected)


def test_block_definitions_values_without_unnamed_args():
    source = """
    ::defblock:alias, blocktype, attr3=value3

    [alias, value1, value2]
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "args": ["value1", "value2"],
            "blocktype": "blocktype",
            "kwargs": {"attr3": "value3"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": None,
            "content": [],
        },
    ]

    with pytest.raises(ParserError):
        _test(source, expected)



from unittest.mock import patch

from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_title_block():
    source = """
    . This is the title
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": None,
            "kwargs": {},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is the title",
                    }
                ],
            },
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


def test_title_without_spaces():
    source = """
    .This is the title
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": None,
            "kwargs": {},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is the title",
                    }
                ],
            },
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


def test_title_with_rich_text():
    source = """
    . This _is_ the title
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": None,
            "kwargs": {},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This ",
                    },
                    {
                        "type": "style",
                        "value": "underscore",
                        "content": {
                            "type": "sentence",
                            "content": [
                                {
                                    "type": "text",
                                    "value": "is",
                                }
                            ],
                        },
                    },
                    {
                        "type": "text",
                        "value": " the title",
                    },
                ],
            },
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


def test_title_block_with_attributes():
    source = """
    . This is the title
    [blocktype,attr1=value1]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"attr1": "value1"},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is the title",
                    }
                ],
            },
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)


@patch("mau.parsers.text_parser.footnote_anchors")
def test_title_block_with_footnote(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = """
    . This is the title[footnote](just some text)
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": None,
            "kwargs": {},
            "secondary_content": [],
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "title": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is the title",
                    },
                    {
                        "content": [
                            {
                                "content": [
                                    {"type": "text", "value": "just some text"}
                                ],
                                "type": "sentence",
                            },
                        ],
                        "defanchor": "defXYZ",
                        "number": 1,
                        "refanchor": "refXYZ",
                        "type": "footnote_ref",
                    },
                ],
            },
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)



# This is a base class for parsers that collects common methods

import functools

from mau.lexers.base_lexer import BaseLexer, Token, TokenTypes, Literal


class TokenError(ValueError):
    """This is used by the parser to exit the context."""


class ConfigurationError(ValueError):
    """This is a configuration error."""


class ParserError(ValueError):
    """This is a detailed parser error"""

    def __init__(self, message=None, context=None):
        super().__init__(message)

        self.context = context


class ExpectedError(ParserError):
    """This exception signals that we were expecting a different token."""


# This makes a function return True instead of
# None when successfully executed, which
# makes the parser loop much easier to
# understand
def parser(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        value = func(*args, **kwargs)
        if value is None:
            return True

    return wrapper_decorator


class BaseParser:
    def __init__(self):
        self.lexer = BaseLexer()

        # This is the position of the current token.
        self.index = -1

        # These are the tokens parsed by the parser.
        self.tokens = []

        # A stack for the parser's state.
        # Currently the state is represented only
        # by the current index in the input tokens.
        self._stack = []

        # These are the nodes created by the parsing.
        self.nodes = []

    @property
    def current_token(self):
        """
        Returns the token being parsed.
        We often need to know which token we are currently
        parsing, but we might already have parsed all
        of them, so this convenience method wraps the
        possible index error.
        """
        if self.index < 0:
            raise ValueError("The parser has no current token")

        try:
            return self.tokens[self.index]
        except IndexError:
            return Token(TokenTypes.EOF)

    def _push(self):
        # Push the current state on the stack.
        self._stack.append(self.index)

    def _pop(self):
        # Get the state from the stack.
        return self._stack.pop()

    def __enter__(self):
        # The parser can be used as a context manager.
        # When we enter a new context we just need to
        # push the current state.
        self._push()

    def __exit__(self, etype, evalue, etrace):
        # This is automatically run when we leave the context.

        # First make sure we leave the stack as it was before
        # we entered the context.
        stacked_index = self._pop()

        # If there was an exception we need to
        # actually backtrace and pretend we didn't
        # do anything. Be cautious and don't get caught.
        if etype:
            # Restore the original position
            self.index = stacked_index

        # If the exception was one of the ones
        # we manage, let's return True so that
        # the exception is not propagated and
        # Python won't be bothered.
        if etype in [TokenError]:
            return True

    def _save(self, node):
        # Store the node.
        self.nodes.append(node)

    def _check_token(self, token, ttype=None, tvalue=None, value_check_function=None):
        # This method performs a test on the current token,
        # figuring out if type and value correspond to those passed
        # as arguments. If type or value are not given they are not
        # tested.
        # If the test is successful the token is returned, otherwise
        # the TokenError exception is raised.
        # The argument value_check_function is a function that
        # can be passed to test the token value and shall return a boolean.

        check_type = ttype if ttype is not None else token.type
        if token.type != check_type:
            raise TokenError(f"Type of token {token} is not {check_type}")

        if tvalue is not None and token.value != tvalue:
            raise TokenError(f"Value of token {token} is not {tvalue}")
        elif (
            value_check_function is not None
            and value_check_function(token.value) is False
        ):
            raise TokenError(f"Value of token {token} didn't pass check")

        return token

    def _parse_functions(self):
        # The parse functions available in this parser
        return []

    def error(self, message=None):
        raise ParserError(message, context=self.lexer.context(self.current_token))

    def put_token(self, token):
        self.tokens.insert(self.index + 1, token)

    def get_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token and advances the index.
        This function returns the next token and then advances the index,
        and can optionally check its type or value (see _check_token).
        The token is stored it in self._current_token.
        """

        if self.index == len(self.tokens):
            return Token(TokenTypes.EOF)

        self.index += 1
        return self._check_token(self.current_token, ttype, tvalue, check)

    def check_current_token(self, ttype, tvalue=None, check=None):
        """
        Just check the type and value of the current token without
        advancing the index.
        """

        return self._check_token(self.current_token, ttype, tvalue, check)

    def peek_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token without advancing the index.
        """

        try:
            token = self.tokens[self.index + 1]
            return self._check_token(token, ttype, tvalue, check)
        except IndexError:
            return Token(TokenTypes.EOF)

    def peek_token_is(self, ttype=None, tvalue=None, check=None):
        """
        Peek a token and check it.
        This works like peek_token, but returns a boolean
        instead of raising an exception.
        """
        try:
            self.peek_token(ttype, tvalue, check)
            return True
        except TokenError:
            return False

    def force_token(self, ttype, tvalue=None):
        """
        Return the next token and advances the index,
        but forces the token to have a specific type
        and optionally a value.
        If the token doesn't match the provided values
        the function raises an ExpectedError
        """
        try:
            return self.get_token(ttype, tvalue)
        except TokenError:
            raise ExpectedError(
                {"expected": Token(ttype, tvalue), "found": self.current_token}
            )

    def collect(self, stop_tokens, preserve_escaped_stop_tokens=False):
        """
        Collect all tokens until one of the stop_tokens pops up.
        Escape tokens "\\" are removed unless
        preserve_escaped_stop_tokens is set to True.
        """
        tokens = []

        while self.peek_token() not in stop_tokens:
            if self.peek_token() == Literal("\\"):
                escape = self.get_token()

                if self.peek_token() not in stop_tokens or preserve_escaped_stop_tokens:
                    tokens.append(escape)

            tokens.append(self.get_token())

        return tokens

    def collect_join(
        self, stop_tokens, join_with="", preserve_escaped_stop_tokens=False
    ):
        """
        Collect all tokens until one of the stop_tokens
        pops up and join them in a single string.
        This works exactly like collect but returns
        a string of tokens joined with the given characters.
        """
        token_values = [
            t.value for t in self.collect(stop_tokens, preserve_escaped_stop_tokens)
        ]

        # EOL tokens have value None, so this removes them
        token_values = [t for t in token_values if t is not None]

        return join_with.join(token_values)

    def load(self, text):
        """
        Load the given text and run the lexer on it.
        """

        # Run the lexer
        self.lexer.process(text)

        # Store the resulting tokens
        self.tokens = self.lexer.tokens

        # Reset the index
        self.index = -1

    def parse(self):
        """
        Run the parser on the lexed tokens.
        """

        # A loop on all lexed tokens until we reach EOF
        while not self.peek_token_is(TokenTypes.EOF):
            result = False

            # Run all parsing functions provided
            # by the parser until one returns
            # a sensible result
            for parse_function in self._parse_functions():
                # The context manager wraps the parse
                # function so that failing functions
                # leave the parsed tokens as they were
                # at the beginning of their execution.
                # The result of a parse function is True
                # only if the function finished properly.
                # If the parse function raised an exception
                # the result is False.
                with self:
                    result = parse_function()

                if result is True:
                    break

            # If we get here and result is still False
            # we didn't find any function to parse the
            # current token.
            if result is False:
                self.error("Cannot parse token")

    def analyse(self, text):
        self.load(text)
        self.parse()

        return self



import copy

from mau.visitors.visitor import Visitor

DEFAULT_TEMPLATES = {
    "block-admonition.md": "{blurb, class: {{ kwargs.class }}}\n**{{ kwargs.label }}**\n\n{{ content }}{/blurb}\n\n",
    "block.md": "{{ content }}",
    "block-source.md": '{% if title %}{caption: "{{ title }}"}\n{% endif %}``` {% if kwargs.language %}{{ kwargs.language }}{% endif %}\n{{ content }}\n```',
    "callout.md": "",
    "class.md": "{{ content }}",
    "command.md": "{{ content }}",
    "document.md": "{{ content }}",
    "footnote_def.md": "[^footnote{{ number }}]: {{ content }}",
    "footnote_ref.md": "[^footnote{{ number }}]",
    "footnotes.md": "{{ entries }}",
    "header.md": "{{ header }} {{ value }}\n",
    "horizontal_rule.md": "* * *",
    "image.md": '{% if alt_text %}{alt: "{{ alt_text }}"}\n{% endif %}![{{ title }}]({{ uri }})\n',
    "inline_image.md": "![{{ alt_text }}]({{ uri }})",
    "link.md": "{% if text %}[{{ text }}]({{ target }}){% else %}<{{ target }}>{% endif %}",
    "list.md": "{{ items }}{% if main_node %}\n{% endif %}",
    "list_item.md": "{% if not main_node %}\n{% endif %}{% if prefix %}{{ prefix }} {% endif %}{{ content }}",
    "paragraph.md": "{{ content }}\n",
    "block-quote.md": "{blurb, icon: quote-right}\n{{ content }}\n{% if kwargs.attribution %}{{ kwargs.attribution }}{% else %}{{ secondary_content }}{% endif %}\n{/blurb}\n\n",
    "sentence.md": "{{ content }}",
    "star.md": "**{{ content }}**",
    "text.md": "{{ value }}",
    "underscore.md": "*{{ content }}*",
    "verbatim.md": "`{{ content }}`",
}


class MarkuaVisitor(Visitor):
    _template_extension = ".md"
    _environment_options = {"keep_trailing_newline": True}

    def __init__(
        self,
        default_templates=None,
        custom_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        default_templates = (
            copy.deepcopy(default_templates)
            if default_templates is not None
            else copy.deepcopy(DEFAULT_TEMPLATES)
        )

        super().__init__(
            default_templates=default_templates,
            custom_templates=custom_templates,
            templates_directory=templates_directory,
            config=config,
            toc=toc,
            footnotes=footnotes,
        )

    def _visit_class(self, node):
        classes = [f".{cls}" for cls in node["classes"]]
        node["classes"] = classes
        return super()._visit_class(node)

    def _visit_link(self, node):
        if node["text"] == node["target"]:
            node["text"] = None
        return node

    def _visit_header(self, node):
        node["header"] = "#" * int(node["level"])
        return node

    def _visit_block_admonition(self, node):
        if node["kwargs"]["class"] == "note":
            node["kwargs"]["class"] = "tip"

        if node["kwargs"]["class"] not in [
            "discussion",
            "error",
            "information",
            "tip",
            "warning",
        ]:
            raise ValueError(
                f"""Admonition {node["kwargs"]["class"]} cannot be converted"""
            )

        return node

    def _visit_block_engine_source(self, node):
        node["kwargs"]["callouts"] = []

        return node

    def _visit_document(self, node):
        self._reducelist(node, ["content"], join_with="\n")
        return node

    def _visit_list_item(self, node, ordered=False):
        mark = "*"

        prefix = None
        if node["content"]["type"] != "list":
            prefix = mark * int(node["level"])

        self._reduce(node, ["content"])
        node["prefix"] = prefix
        return node

    def _visit_list(self, node, ordered=False):
        node["items"] = "".join(
            [self.visit(i, ordered=node["ordered"]) for i in node["items"]]
        )
        return node

    def _visit_command_footnotes(self, node):
        footnotes_node = self.footnotes.asdict()
        self._reducelist(footnotes_node, ["entries"], join_with="")

        return footnotes_node



import textwrap


def dedent(text):
    return textwrap.dedent(text).strip()


def remove_indentation(text):
    # This removes indentation and newlines completely
    return "".join([i.strip() for i in textwrap.dedent(text).split("\n")])


def listasdict(nodes):
    return [i.asdict() for i in nodes]


def init_parser_factory(parser_class):
    def _init_parser(text, *args, **kwargs):
        p = parser_class(*args, **kwargs)

        p.load(text)

        return p

    return _init_parser


def parser_test_factory(parser_class, *args, **kwds):

    init_parser = init_parser_factory(parser_class)

    def _test(source, expected):
        p = init_parser(textwrap.dedent(source), *args, **kwds)
        p.parse()

        assert listasdict(p.nodes) == expected

        return p

    return _test


def init_ast_factory(parser_class):
    def _init_ast(text, *args, **kwargs):
        init_parser = init_parser_factory(parser_class)

        p = init_parser(text, *args, **kwargs)
        p.parse()

        return listasdict(p.nodes)

    return _init_ast


def visitlist_factory(visitor_class):
    def _visitlist(nodes, *args, **kwargs):
        v = visitor_class(*args, **kwargs)

        return [v.visit(node) for node in nodes]

    return _visitlist


def ast_test_factory(parser_class, visitor_class):
    init_ast = init_ast_factory(parser_class)
    visitlist = visitlist_factory(visitor_class)

    def _test(source, expected):
        ast = init_ast(source)
        result = visitlist(ast)

        assert result == expected

    return _test



from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_admonition():
    source = """
    [admonition,someclass,someicon,somelabel]
    ----
    Content
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": "admonition",
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "content": [{"type": "text", "value": "Content"}],
                        "type": "sentence",
                    },
                }
            ],
            "secondary_content": [],
            "title": None,
            "classes": [],
            "engine": "default",
            "preprocessor": "none",
            "args": [],
            "kwargs": {"class": "someclass", "icon": "someicon", "label": "somelabel"},
        },
    ]

    _test(source, expected)


def test_parse_block_quote():
    source = """
    [quote,"Star Wars, 1977"]
    ----
    Learn about the Force, Luke.
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "quote",
            "classes": [],
            "engine": "default",
            "kwargs": {"attribution": "Star Wars, 1977"},
            "preprocessor": "none",
            "title": None,
            "secondary_content": [],
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Learn about the Force, Luke."}
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_parse_block_quote_attribution_in_secondary_content():
    source = """
    [quote]
    ----
    Learn about the Force, Luke.
    ----
    _Star Wars_, 1977
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "quote",
            "classes": [],
            "engine": "default",
            "kwargs": {"attribution": None},
            "preprocessor": "none",
            "title": None,
            "secondary_content": [
                {
                    "args": [],
                    "content": {
                        "content": [
                            {
                                "content": {
                                    "content": [{"type": "text", "value": "Star Wars"}],
                                    "type": "sentence",
                                },
                                "type": "style",
                                "value": "underscore",
                            },
                            {"type": "text", "value": ", 1977"},
                        ],
                        "type": "sentence",
                    },
                    "kwargs": {},
                    "type": "paragraph",
                },
            ],
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Learn about the Force, Luke."}
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


# def test_parse_block_quote_attribution_in_secondary_content_supports_styles():
#     source = """
#     [quote,"_Star Wars_, 1977"]
#     ----
#     Learn about the Force, Luke.
#     ----
#     """

#     expected = [
#         {
#             "type": "quote",
#             "attribution": {
#                 "content": [
#                     {
#                         "content": {
#                             "content": [{"type": "text", "value": "Star Wars"}],
#                             "type": "sentence",
#                         },
#                         "type": "style",
#                         "value": "underscore",
#                     },
#                     {"type": "text", "value": ", 1977"},
#                 ],
#                 "type": "sentence",
#             },
#             "kwargs": {},
#             "content": [
#                 {
#                     "type": "paragraph",
#                     "args": [],
#                     "kwargs": {},
#                     "content": {
#                         "type": "sentence",
#                         "content": [
#                             {"type": "text", "value": "Learn about the Force, Luke."}
#                         ],
#                     },
#                 }
#             ],
#         }
#     ]

#     _test(source, expected)
